// LT ------------------------------------------------------------------------//
//
// cMatrix
//
// General Matrix stuff
//
// IMPORTANT!
// 
// Matricies are laid out following the OpenGL spec, as:
//	[00, 04, 08, 12]
//	[01, 05, 09, 13]
//	[02, 06, 10, 14]
//	[03, 07, 11, 15]
//
// This means accessing an element using [][] is the other way around to the
// standard C array convention - use [column][row] rather than [row][column].
// 
// Piran Tremethick
// 08/02/06
//----------------------------------------------------------------------------//

#ifndef _LT_MATRIX_H_
#define _LT_MATRIX_H_

// Standard headers
#include <ostream>

// LT headers
#include "LT_vector.h"

// Forward declare quaternions as they're used here, but quaternion also
// needs to know about matrix... mmm, cyclic.
class cQuaternion;

class cMatrix4D
{
public:
	union{
		float _m[16];
		float m[4][4];
	};

	cMatrix4D();
	cMatrix4D(const cMatrix4D& in);
	cMatrix4D(const float* in);	

	inline const float* operator[] (unsigned int col) const
	{
		return m[col];
	}

	inline cMatrix4D& operator= (const cMatrix4D& in)
	{
		m[0][0] = in.m[0][0];
		m[0][1] = in.m[0][1];
		m[0][2] = in.m[0][2];
		m[0][3] = in.m[0][3];

		m[1][0] = in.m[1][0];
		m[1][1] = in.m[1][1];
		m[1][2] = in.m[1][2];
		m[1][3] = in.m[1][3];

		m[2][0] = in.m[2][0];
		m[2][1] = in.m[2][1];
		m[2][2] = in.m[2][2];
		m[2][3] = in.m[2][3];

		m[3][0] = in.m[3][0];
		m[3][1] = in.m[3][1];
		m[3][2] = in.m[3][2];
		m[3][3] = in.m[3][3];

		return *this;
	}

	inline cMatrix4D operator* (const cMatrix4D& in) const
	{		
		cMatrix4D t;

		t.m[0][0] = m[0][0]*in.m[0][0] + m[1][0]*in.m[0][1] + m[2][0]*in.m[0][2] + m[3][0]*in.m[0][3];
		t.m[0][1] = m[0][1]*in.m[0][0] + m[1][1]*in.m[0][1] + m[2][1]*in.m[0][2] + m[3][1]*in.m[0][3];
		t.m[0][2] = m[0][2]*in.m[0][0] + m[1][2]*in.m[0][1] + m[2][2]*in.m[0][2] + m[3][2]*in.m[0][3];
		t.m[0][3] = m[0][3]*in.m[0][0] + m[1][3]*in.m[0][1] + m[2][3]*in.m[0][2] + m[3][3]*in.m[0][3];
		
		t.m[1][0] = m[0][0]*in.m[1][0] + m[1][0]*in.m[1][1] + m[2][0]*in.m[1][2] + m[3][0]*in.m[1][3];
		t.m[1][1] = m[0][1]*in.m[1][0] + m[1][1]*in.m[1][1] + m[2][1]*in.m[1][2] + m[3][1]*in.m[1][3];
		t.m[1][2] = m[0][2]*in.m[1][0] + m[1][2]*in.m[1][1] + m[2][2]*in.m[1][2] + m[3][2]*in.m[1][3];
		t.m[1][3] = m[0][3]*in.m[1][0] + m[1][3]*in.m[1][1] + m[2][3]*in.m[1][2] + m[3][3]*in.m[1][3];
		
		t.m[2][0] = m[0][0]*in.m[2][0] + m[1][0]*in.m[2][1] + m[2][0]*in.m[2][2] + m[3][0]*in.m[2][3];
		t.m[2][1] = m[0][1]*in.m[2][0] + m[1][1]*in.m[2][1] + m[2][1]*in.m[2][2] + m[3][1]*in.m[2][3];
		t.m[2][2] = m[0][2]*in.m[2][0] + m[1][2]*in.m[2][1] + m[2][2]*in.m[2][2] + m[3][2]*in.m[2][3];
		t.m[2][3] = m[0][3]*in.m[2][0] + m[1][3]*in.m[2][1] + m[2][3]*in.m[2][2] + m[3][3]*in.m[2][3];
		
		t.m[3][0] = m[0][0]*in.m[3][0] + m[1][0]*in.m[3][1] + m[2][0]*in.m[3][2] + m[3][0]*in.m[3][3];
		t.m[3][1] = m[0][1]*in.m[3][0] + m[1][1]*in.m[3][1] + m[2][1]*in.m[3][2] + m[3][1]*in.m[3][3];
		t.m[3][2] = m[0][2]*in.m[3][0] + m[1][2]*in.m[3][1] + m[2][2]*in.m[3][2] + m[3][2]*in.m[3][3];
		t.m[3][3] = m[0][3]*in.m[3][0] + m[1][3]*in.m[3][1] + m[2][3]*in.m[3][2] + m[3][3]*in.m[3][3];

		return t;
	}

	inline cVector3D operator* (const cVector3D& in) const
	{
		// NOTE - assuming the in vector is actually a homogeneous vector with
		// a w component of 1.0. This is a little lazy, but saves having to 
		// have a seperate 4D vector type when most of the time w will be
		// ignored. 
		// Will sort out a 4D vector type when everything's working.
		
		cVector3D t;

		t.x = m[0][0]*in.x + m[1][0]*in.y + m[2][0]*in.z + m[3][0];
		t.y = m[0][1]*in.x + m[1][1]*in.y + m[2][1]*in.z + m[3][1];
		t.z = m[0][2]*in.x + m[1][2]*in.y + m[2][2]*in.z + m[3][2];

		return t;
	}
	
	inline cMatrix4D operator+ (const cMatrix4D& in) const
	{
		cMatrix4D t;

		t.m[0][0] = m[0][0] + in.m[0][0];
		t.m[0][1] = m[0][1] + in.m[0][1];
		t.m[0][2] = m[0][2] + in.m[0][2];
		t.m[0][3] = m[0][3] + in.m[0][3];

		t.m[1][0] = m[1][0] + in.m[1][0];
		t.m[1][1] = m[1][1] + in.m[1][1];
		t.m[1][2] = m[1][2] + in.m[1][2];
		t.m[1][3] = m[1][3] + in.m[1][3];

		t.m[2][0] = m[2][0] + in.m[2][0];
		t.m[2][1] = m[2][1] + in.m[2][1];
		t.m[2][2] = m[2][2] + in.m[2][2];
		t.m[2][3] = m[2][3] + in.m[2][3];

		t.m[3][0] = m[3][0] + in.m[3][0];
		t.m[3][1] = m[3][1] + in.m[3][1];
		t.m[3][2] = m[3][2] + in.m[3][2];
		t.m[3][3] = m[3][3] + in.m[3][3];

		return t;
	}

	inline cMatrix4D operator- (const cMatrix4D& in) const
	{
		cMatrix4D t;

		t.m[0][0] = m[0][0] - in.m[0][0];
		t.m[0][1] = m[0][1] - in.m[0][1];
		t.m[0][2] = m[0][2] - in.m[0][2];
		t.m[0][3] = m[0][3] - in.m[0][3];

		t.m[1][0] = m[1][0] - in.m[1][0];
		t.m[1][1] = m[1][1] - in.m[1][1];
		t.m[1][2] = m[1][2] - in.m[1][2];
		t.m[1][3] = m[1][3] - in.m[1][3];

		t.m[2][0] = m[2][0] - in.m[2][0];
		t.m[2][1] = m[2][1] - in.m[2][1];
		t.m[2][2] = m[2][2] - in.m[2][2];
		t.m[2][3] = m[2][3] - in.m[2][3];

		t.m[3][0] = m[3][0] - in.m[3][0];
		t.m[3][1] = m[3][1] - in.m[3][1];
		t.m[3][2] = m[3][2] - in.m[3][2];
		t.m[3][3] = m[3][3] - in.m[3][3];

		return t;
	}
	
	inline void translate(const cVector3D& t)
	{
		m[3][0] += t.x;
		m[3][1] += t.y;
		m[3][2] += t.z;
	}
		
	void setIdentity(void);
	void setFromEuler(float rx, float ry, float rz);
	void setFromAxisAngle(cVector3D axis, float ang);
	void setFromQuat(const cQuaternion& q);
	
	void setTransform(const cVector3D& t);
	void setTransform(float xx, float yy, float zz);
	void setRotation(const cMatrix4D& r);
	cVector3D getTransform(void);
};

std::ostream& operator<< (std::ostream& os, const cMatrix4D& in);

#endif //_LT_MATRIX_H_
