#include <iostream>
#include <string>
#include <sstream>

#include "LT_app.h"
#include "LT_sdl.h"
#include "LT_xml.h"
#include "LT_math.h"
#include "LT_quaternion.h"
#include "LT_matrix.h"

void cLTApp::Setup(unsigned int w, unsigned int h)
{
	const SDL_VideoInfo* info;

	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) != 0)
	{
		ExitError(SDL_GetError());
	};
	
	info = SDL_GetVideoInfo();
	
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	
	if(SDL_SetVideoMode(w, h, info->vfmt->BitsPerPixel, SDL_OPENGL) == 0)
	{
		ExitError(SDL_GetError());
	};

	glClearColor(0.5f, 0.5f, 0.6f, 1.0f);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, static_cast<float>(w)/h, 0.1f, 100.0f);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	
	mWinWidth = w;
	mWinHeight = h;
	mDrawMode = 1;
	mLeftMouseButton = false;
	mRightMouseButton = false;
	mActive = true;
	mPaused = false;
	mTurnLeft = false;
	mTurnRight = false;
	mWalk = false;

	mFPS = 0;
	mCurrentTime = SDL_GetTicks();
	mLastTime = SDL_GetTicks();
}

void cLTApp::SetupCamera(float xx, float yy, float zz, float dist, int heading, int elevation)
{
	mCamera.Setup(xx, yy, zz, dist, heading, elevation);
}

void cLTApp::LoadActor(const char* fname)
{
	LoadActorFromXML(fname, mActor);
	mActor.SetConstraints();
	mActor.UpdateArmature();
}

void cLTApp::Go(void)
{
	while(mActive)
	{
		HandleEvents();
        Update();
		Draw();
	}	

	SDL_Quit();
}

void cLTApp::HandleEvents(void)
{
	SDL_Event evt;
	
	while(SDL_PollEvent(&evt))
	{
		switch(evt.type)
		{
			case SDL_QUIT:
				mActive = 0;
				break;
			case SDL_KEYDOWN:
				if(evt.key.keysym.sym == 'a')
				{
					mTurnLeft = !mTurnLeft;
				}
				else if(evt.key.keysym.sym == 'd')
				{
					mTurnRight = !mTurnRight;
				}
				else if(evt.key.keysym.sym == 'w')
				{
					mWalk = !mWalk;
				}
				break;
			case SDL_KEYUP:
				if(evt.key.keysym.sym == 'p')
				{
					mPaused = !mPaused;
				}
				else if(evt.key.keysym.sym == 'a')
				{
					mTurnLeft = !mTurnLeft;
				}
				else if(evt.key.keysym.sym == 'd')
				{
					mTurnRight = !mTurnRight;
				}
				else if(evt.key.keysym.sym == 'w')
				{
					mWalk = !mWalk;
				}
				else if(evt.key.keysym.sym == 't')
				{
					if(mDrawMode == 2)
						mDrawMode = 1;
					else
						++mDrawMode;
				}
				else if(evt.key.keysym.sym == 'c')
				{
					if(mActor.mWalkCycle == 0)
						mActor.mWalkCycle = 2;
					else
						mActor.mWalkCycle = 0;
				}
				break;
			case SDL_MOUSEBUTTONDOWN:
				if(evt.button.button == SDL_BUTTON_LEFT)
					mLeftMouseButton = true;
				else if(evt.button.button == SDL_BUTTON_RIGHT)
					mRightMouseButton = true;
				break;
			case SDL_MOUSEBUTTONUP:
				if(evt.button.button == SDL_BUTTON_LEFT)
					mLeftMouseButton = false;
				else if(evt.button.button == SDL_BUTTON_RIGHT)
					mRightMouseButton = false;
			case SDL_MOUSEMOTION:
				if(mLeftMouseButton)
				{
					mCamera.Turn(evt.motion.xrel, evt.motion.yrel);
				}
				break;
			default:
				break;
		}
	}
}

void cLTApp::Update(void)
{
	static std::stringstream strm;
	static std::string caption;
	static unsigned int dt = 0;

	while(mCurrentTime <= mLastTime)
	{
		mCurrentTime = SDL_GetTicks();
	}
	
	dt = mCurrentTime - mLastTime;
	if((mCurrentTime - mSecondStartTime) > 1000)
	{
		mSecondStartTime = mCurrentTime;
		strm<<mFPS;
		strm>>caption;
		strm.clear();

		caption = "FPS: " + caption;
		SDL_WM_SetCaption(caption.c_str(), NULL);
		mFPS = 0;
	}

	if(!mPaused)
	{
		if(mWalk)
		{
			if(mTurnLeft)
			{
				mActor.mHeading += 0.1*dt;
			}
			if(mTurnRight)
			{
				mActor.mHeading -= 0.1*dt;
			}
			if(mActor.mNextAction != mActor.mWalkCycle)
			{
				mActor.mNextAction = mActor.mWalkCycle;
			}
		}
		else
		{
			if(mActor.mNextAction != 1)
			{
				mActor.mNextAction = 1;
			}
		}
		mActor.Animate(dt);
		mCamera.SetTarget(mActor.mArmature.mRootBone->mWorldTransform.getTransform());
	}
	mCamera.Update();

	++mFPS;
	mLastTime = SDL_GetTicks();
}

void cLTApp::Draw(void)
{	
	cVector3D root;

	glClearColor(0.5f, 0.5f, 0.6f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	gluLookAt(mCamera.mPosition.x, mCamera.mPosition.y, mCamera.mPosition.z,
				mCamera.mTarget.x, mCamera.mTarget.y, mCamera.mTarget.z,
				0.0, 1.0, 0.0);
	
	glColor3f(0.0f, 0.0f, 0.0f);
	
	glBegin(GL_LINES);
	for(int i=-100; i<101; ++i)
	{
		glVertex3f((GLfloat)i, 0.0, 100.0);
		glVertex3f((GLfloat)i, 0.0, -100.0);
		
		glVertex3f(100.0, 0.0, (GLfloat)i);
		glVertex3f(-100.0, 0.0, (GLfloat)i);
	}
	
	glEnd();
	
	mActor.DrawArmature(mDrawMode);

	root = mActor.mArmature.mRootBone->mWorldTransform.getTransform();
	glColor3f(0.0, 1.0, 0.0);
	glBegin(GL_LINES);
	glVertex3f(root.x, root.y, root.z);
	glVertex3f(root.x+mActor.mLeftFoot.mDir.x, root.y+mActor.mLeftFoot.mDir.y, root.z+mActor.mLeftFoot.mDir.z);
	glEnd();

	glPointSize(4.0);
	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_POINTS);
	glVertex3f(mActor.mRightFoot.mLastPosition.x, mActor.mRightFoot.mLastPosition.y, mActor.mRightFoot.mLastPosition.z);
	glVertex3f(mActor.mLeftFoot.mLastPosition.x, mActor.mLeftFoot.mLastPosition.y, mActor.mLeftFoot.mLastPosition.z);
	glEnd();

	glFlush();
	SDL_GL_SwapBuffers();
}

void cLTApp::ExitError(const char* err)
{
	std::string error("EXIT ON ERROR: ");
	error += err;

	SDL_Quit();

#ifdef WIN32
	MessageBox(HWND_DESKTOP, error.c_str(), "Error", MB_OK|MB_ICONEXCLAMATION);
	exit(1);
#else
	std::cout<<error<<std::endl;
	exit(1);
#endif

}
