// LT ------------------------------------------------------------------------//
//
// cMatrix
//
// General Matrix stuff
//
// IMPORTANT!
// 
// Matricies are laid out following the OpenGL spec, as:
//	[00, 04, 08, 12]
//	[01, 05, 09, 13]
//	[02, 06, 10, 14]
//	[03, 07, 11, 15]
//
// This means accessing an element using [][] is the other way around to the
// standard C array convention - use [column][row] rather than [row][column].
// 
// Piran Tremethick
// 08/02/06
//----------------------------------------------------------------------------//

// Standard headers
#include <ostream>

// LT headers
#include "LT_math.h"
#include "LT_matrix.h"
#include "LT_quaternion.h"

// Constructor
// Set to identity by default
//
cMatrix4D::cMatrix4D()
{
	_m[0] = 1.0; _m[4] = 0.0; _m[8] = 0.0;  _m[12] = 0.0;	
	_m[1] = 0.0; _m[5] = 1.0; _m[9] = 0.0;  _m[13] = 0.0;
	_m[2] = 0.0; _m[6] = 0.0; _m[10] = 1.0; _m[14] = 0.0;
	_m[3] = 0.0; _m[7] = 0.0; _m[11] = 0.0; _m[15] = 1.0;
}

// Copy constructor
//
cMatrix4D::cMatrix4D(const cMatrix4D& in)
{
	m[0][0] = in.m[0][0];
	m[0][1] = in.m[0][1];
	m[0][2] = in.m[0][2];
	m[0][3] = in.m[0][3];

	m[1][0] = in.m[1][0];
	m[1][1] = in.m[1][1];
	m[1][2] = in.m[1][2];
	m[1][3] = in.m[1][3];

	m[2][0] = in.m[2][0];
	m[2][1] = in.m[2][1];
	m[2][2] = in.m[2][2];
	m[2][3] = in.m[2][3];

	m[3][0] = in.m[3][0];
	m[3][1] = in.m[3][1];
	m[3][2] = in.m[3][2];
	m[3][3] = in.m[3][3];
}

// Constructor
// Set from a float array
//
cMatrix4D::cMatrix4D(const float* a)
{
	_m[0] = a[0]; _m[4] = a[4]; _m[8] = a[8];   _m[12] = a[12];	
	_m[1] = a[1]; _m[5] = a[5]; _m[9] = a[9];   _m[13] = a[13];
	_m[2] = a[2]; _m[6] = a[6]; _m[10] = a[10]; _m[14] = a[14];
	_m[3] = a[3]; _m[7] = a[7]; _m[11] = a[11]; _m[15] = a[15];
}

// setIdentity
// Make this an identity matrix
//
void cMatrix4D::setIdentity(void)
{
	m[0][0] = 1.0;
	m[0][1] = 0.0;
	m[0][2] = 0.0;
	m[0][3] = 0.0;
	
	m[1][0] = 0.0;
	m[1][1] = 1.0;
	m[1][2] = 0.0;
	m[1][3] = 0.0;
	
	m[2][0] = 0.0;
	m[2][1] = 0.0;
	m[2][2] = 1.0;
	m[2][3] = 0.0;
	
	m[3][0] = 0.0;
	m[3][1] = 0.0;
	m[3][2] = 0.0;
	m[3][3] = 1.0;
}

// setFromEuler
// params: float angles around X, Y, Z axis
// make this a rotation matrix from X, Y, Z Euler angles
//
void cMatrix4D::setFromEuler(float rx, float ry, float rz)
{
	float cosZ = Math::cos(rz);
	float cosY = Math::cos(ry);
	float cosX = Math::cos(rx);
	float sinZ = Math::sin(rz);
	float sinY = Math::sin(ry);
	float sinX = Math::sin(rx);
	
	m[0][0] = cosZ*cosY;
	m[0][1] = sinZ*cosY;
	m[0][2] = -sinY;
	m[0][3] = 0.0;
	
	m[1][0] = -sinZ*cosX + cosZ*sinY*sinX;
	m[1][1] = cosZ*cosX + sinZ*sinY*sinX;
	m[1][2] = cosY*sinX;
	m[1][3] = 0.0;
	
	m[2][0] = -sinZ*-sinX + cosZ*sinY*cosX;
	m[2][1] = cosZ*-sinX + sinZ*sinY*cosX;
	m[2][2] = cosY*cosX;
	m[2][3] = 0.0;
	
	/*m[3][0] = 0.0;
	m[3][1] = 0.0;
	m[3][2] = 0.0;
	m[3][3] = 1.0;*/
}

// setFromQuat
// params: a unit quaternion representation of a rotation
// make this a rotation matrix from a quaternion
//
void cMatrix4D::setFromQuat(const cQuaternion& q)
{
	m[0][0] = 1.0 - 2*Math::sqr(q.y) - 2*Math::sqr(q.z);
	m[0][1] = 2*q.x*q.y + 2*q.w*q.z;
	m[0][2] = 2*q.x*q.z - 2*q.w*q.y;
	m[0][3] = 0.0;
	
	m[1][0] = 2*q.x*q.y - 2*q.w*q.z;
	m[1][1] = 1.0 - 2*Math::sqr(q.x) - 2*Math::sqr(q.z);
	m[1][2] = 2*q.y*q.z + 2*q.w*q.x;
	m[1][3] = 0.0;
	
	m[2][0] = 2*q.x*q.z + 2*q.w*q.y;
	m[2][1] = 2*q.y*q.z - 2*q.w*q.x;
	m[2][2] = 1.0 - 2*Math::sqr(q.x) - 2*Math::sqr(q.y);
	m[2][3] = 0.0;
	
}

// setFromAxisAngle
// params: an axis vector and a float angle
// make this matrix a rotation matrix from a rotation axis and an angle
// around said axis.
//
void cMatrix4D::setFromAxisAngle(cVector3D axis, float ang)
{
	float sinA = Math::sin(ang);
	float invcosA = 1-Math::cos(ang);
	
	axis.normalize();
	
	m[0][0] = 1+invcosA * (axis.x*axis.x-1);
	m[0][1] = axis.z * sinA + invcosA*axis.x*axis.y;
	m[0][2] = -axis.y*sinA + invcosA*axis.x*axis.z;
	
	m[1][0] = -axis.z*sinA + invcosA*axis.x*axis.y;
	m[1][1] = 1+invcosA * (axis.y*axis.y-1);
	m[1][2] = axis.x*sinA + invcosA*axis.y*axis.z;
	
	m[2][0] = axis.y*sinA + invcosA*axis.x*axis.z;
	m[2][1] = -axis.x*sinA + invcosA*axis.y*axis.z;
	m[2][2] = 1 + invcosA * (axis.z*axis.z-1);
}

// setTransform
// params: a vector
// set just the transform component of this matrix
//
void cMatrix4D::setTransform(const cVector3D& t)
{
	m[3][0] = t.x;	
	m[3][1] = t.y;
	m[3][2] = t.z;
	m[3][3] = 1.0;
}

// setTransform
// params: float X, Y, Z positions
// set just the transform component of this matrix from X, Y, Z values
//
void cMatrix4D::setTransform(float xx, float yy, float zz)
{
	m[3][0] = xx;	
	m[3][1] = yy;
	m[3][2] = zz;
	m[3][3] = 1.0;
}

// setRotation
// params: a matrix
// set just the 3x3 rotation sub-matrix from the passed in matrix
//
void cMatrix4D::setRotation(const cMatrix4D& r)
{
	m[0][0] = r.m[0][0];
	m[0][1] = r.m[0][1];
	m[0][2] = r.m[0][2];
	
	m[1][0] = r.m[1][0];
	m[1][1] = r.m[1][1];
	m[1][2] = r.m[1][2];
	
	m[2][0] = r.m[2][0];
	m[2][1] = r.m[2][1];
	m[2][2] = r.m[2][2];
}

// getTransform
// return just the transform component of this matrix
//
cVector3D cMatrix4D::getTransform(void)
{
	cVector3D t;
	
	t.x = m[3][0];	
	t.y = m[3][1];
	t.z = m[3][2];
	
	return t;
}

// Output stream operator
//
std::ostream& operator<< (std::ostream& os, const cMatrix4D& in)
{
	os<<"Matrix4D: ["<<in._m[0]<<", "<<in._m[4]<<", "<<in._m[8]<<", "<<in._m[12]<<" ]"<<std::endl;
	os<<"          ["<<in._m[1]<<", "<<in._m[5]<<", "<<in._m[9]<<", "<<in._m[13]<<" ]"<<std::endl;
	os<<"          ["<<in._m[2]<<", "<<in._m[6]<<", "<<in._m[10]<<", "<<in._m[14]<<" ]"<<std::endl;
	os<<"          ["<<in._m[3]<<", "<<in._m[7]<<", "<<in._m[11]<<", "<<in._m[15]<<" ]";

	return os;
}
