// LT ------------------------------------------------------------------------//
//
// cQuaternion
// General Quaternion stuff
//
// Piran Tremethick
// 08/02/06
//----------------------------------------------------------------------------//

// Standard headers
#include <iostream>
#include <ostream>

// LT headers
#include "LT_quaternion.h"

// Constructor
// Identity
cQuaternion::cQuaternion()
{
	x = y = z = 0.0;
	w = 1.0;	
}

// Copy Constructor
//
cQuaternion::cQuaternion(const cQuaternion& q)
{
	x = q.x;	
	y = q.y;
	z = q.z;
	w = q.w;
}

// setIdentity
//
void cQuaternion::setIdentity(void)
{
	x = y = z = 0.0;
	w = 1.0;	
}

// setFromEuler
// params: float X, Y, Z angles
// make this quaternion represent the rotation given by X, Y, Z
// angles.
//
void cQuaternion::setFromEuler(float xx, float yy, float zz)
{
	cQuaternion qx, qy, qz, t;
		
	qx.w = Math::cos(xx/2);
	qx.x = Math::sin(xx/2);
	qx.y = 0.0;
	qx.z = 0.0;
	
	qy.w = Math::cos(yy/2);
	qy.x = 0.0;
	qy.y = Math::sin(yy/2);
	qy.z = 0.0;
	
	qz.w = Math::cos(zz/2);
	qz.x = 0.0;
	qz.y = 0.0;
	qz.z = Math::sin(zz/2);
	
	t = qx*qy*qz;
	
	x = t.x;
	y = t.y;
	z = t.z;
	w = t.w;
	
}

// setFromMatrix
// params: 4x4 transformation matrix
// make this quaternion represent the rotation held in the rotation part
// of the passed in matrix.
//
void cQuaternion::setFromMatrix(const cMatrix4D& m)
{
	float t = m[0][0] + m[1][1] + m[2][2] + 1.0;
	float s;
	
	if( t > Math::ALMOSTZERO)
	{
		s = 0.5 / Math::sqrt(t);
		w = 0.25 / s;
		x = (m[1][2] - m[2][1]) * s;
		y = (m[2][0] - m[0][2]) * s;
		z = (m[0][1] - m[1][0]) * s;
	}
	else
	{
		if(m[0][0] > m[1][1] && m[0][0] > m[2][2])
		{
			s = 2.0 * Math::sqrt(1.0 + m[0][0] - m[1][1] - m[2][2]);
			x = 0.25 * s;
			y = (m[1][0] + m[0][1]) / s;
			z = (m[2][0] + m[0][2]) / s;
			w = (m[2][1] - m[1][2]) / s;	
		}
		else if(m[1][1] >  m[2][2])
		{
			s = 2.0 * Math::sqrt(1.0 + m[1][1] - m[0][0] - m[2][2]);
			x = (m[1][0] + m[0][1]) / s;
			y = 0.25 * s;
			z = (m[2][1] + m[1][2]) / s;
			w = (m[2][0] - m[0][2]) / s;	
		}
		else
		{
			s = 2.0 * Math::sqrt(1.0 + m[2][2] - m[0][0] - m[1][1]);
			x = (m[2][0] + m[0][2]) / s;
			y = (m[2][1] + m[1][2]) / s;
			z = 0.25 * s;
			w = (m[1][0] - m[0][1]) / s;	
		}
	}
}

//cQuaternion cQuaternion::slerp(float t, const cQuaternion& s, const cQuaternion& d);

// nlerp
// params: float time, start quaternion, destination quaternion
// perform a normalized linear interpolation between the start and
// destination rotation quaternions.
// Although slerp is normally used, because rotations are represented by unit
// quaternions the result of a normalized linear interpolation is actually the
// same arc as if using slerp. So using the faster nlerp. The disadvantage is
// that the speed of the rotation is not constant - but isn't really noticable.
//
cQuaternion cQuaternion::nlerp(float t, const cQuaternion& s, const cQuaternion& d)
{
	// TODO: make robust, for now assuming quaternions are less than
	// 90 degrees apart.

	cQuaternion res;
	
	res.x = s.x + t*(d.x-s.x);
	res.y = s.y + t*(d.y-s.y);
	res.z = s.z + t*(d.z-s.z);
	res.w = s.w + t*(d.w-s.w);
	
	res.normalize();
	
	return res;
}

// Output stream operator
//
std::ostream& operator<< (std::ostream& os, const cQuaternion& in)
{
	os<<"Quaternion: ["<<in.x<<", "<<in.y<<", "<<in.z<<"], "<<in.w<<".";
	return os;	
}
